#ifndef BINVEC_H
#define BINVEC_H

#include <assert.h>
#include <limits.h>
#include <memory.h>
#include "util.h"

typedef unsigned long long uint64;

// define BinVec128, BinVec384, BinVec1280, BinVec1536
__declspec( align( 16 ) ) struct BinVec128
   {
   BinVec128() { _bits[ 1 ] = _bits[ 0 ] = 0; }
   int GetLength() const { return sizeof( _bits ) * 8; }
   void Or( BinVec128 const& a, BinVec128 const& b )
      {
      _bits[ 0 ] = a._bits[ 0 ] | b._bits[ 0 ];
      _bits[ 1 ] = a._bits[ 1 ] | b._bits[ 1 ];
      }
   void And( BinVec128 const& a, BinVec128 const& b )
      {
      _bits[ 0 ] = a._bits[ 0 ] & b._bits[ 0 ];
      _bits[ 1 ] = a._bits[ 1 ] & b._bits[ 1 ];
      }
   void SetAllBitsTo1() { memset( _bits, 0xFF, sizeof( _bits ) ); }
   void SetBit( int bit )
      {
      assert( bit >= 0  &&  bit < GetLength() );
      _bits[ bit >> 6 ] |= (static_cast< uint64 >( 1 ) << (bit & 0x3F));
      }
   bool IsBitSet( int bit ) const
      {
      assert( bit >= 0  &&  bit < GetLength() );
      return (_bits[ bit >> 6 ] & (static_cast< uint64 >( 1 ) << (bit & 0x3F))) != 0;
      }
   int GetLowestBitSet() const
      {
      uint64 data = _bits[ 0 ];
      int bit = 0;
      if ( data == 0 )
         {
         data = _bits[ 1 ];
         bit = 64;
         }
      __asm
         {
         bsf rax, data
         add bit, eax
         }
      return bit;
      }
   bool AreAllBitsZero() const
      {
      return _bits[ 0 ] == 0  &&  _bits[ 1 ] == 0;
      }
   bool operator==( BinVec128 const& rhs ) const
      {
      return _bits[ 0 ] == rhs._bits[ 0 ]  &&  _bits[ 1 ] == rhs._bits[ 1 ];
      }
   void Reset()
      {
      _bits[ 1 ] = _bits[ 0 ] = 0;
      }
   int GetWeightAndNot( BinVec128 const& rhs ) const
      {
      int weight;
      __asm
         {
         mov    rdi, rhs
         mov    rsi, this
         movdqa xmm0, [rdi]
         pandn  xmm0, [rsi]
         movq   rax, xmm0
         pextrq rbx, xmm0, 1
         popcnt rax, rax
         popcnt ebx, ebx
         add    eax, ebx
         mov    weight, eax
         }
      return weight;
      }
   
   uint64 _bits[ 2 ];
   };

__declspec( align( 16 ) ) struct BinVec384
   {
   BinVec384() { memset( _bits, 0, sizeof( _bits ) ); }
   int GetLength() const { return sizeof( _bits ) * 8; }
   void And( BinVec384 const& a, BinVec384 const& b )
      {
      _bits[ 0 ] = a._bits[ 0 ] & b._bits[ 0 ];
      _bits[ 1 ] = a._bits[ 1 ] & b._bits[ 1 ];
      _bits[ 2 ] = a._bits[ 2 ] & b._bits[ 2 ];
      _bits[ 3 ] = a._bits[ 3 ] & b._bits[ 3 ];
      _bits[ 4 ] = a._bits[ 4 ] & b._bits[ 4 ];
      _bits[ 5 ] = a._bits[ 5 ] & b._bits[ 5 ];
      }
   void SetAllBitsTo1() { memset( _bits, 0xFF, sizeof( _bits ) ); }
   void SetBit( int bit )
      {
      assert( bit >= 0  &&  bit < GetLength() );
      _bits[ bit >> 6 ] |= (static_cast< uint64 >( 1 ) << (bit & 0x3F));
      }
   bool IsBitSet( int bit ) const
      {
      assert( bit >= 0  &&  bit < GetLength() );
      return (_bits[ bit >> 6 ] & (static_cast< uint64 >( 1 ) << (bit & 0x3F))) != 0;
      }
   bool AreAllBitsZero() const
      {
      return _bits[ 0 ] == 0  &&  _bits[ 1 ] == 0  &&
             _bits[ 2 ] == 0  &&  _bits[ 3 ] == 0  &&
             _bits[ 4 ] == 0  &&  _bits[ 5 ] == 0;
      }
   int GetLowestBitSet() const
      {
      for ( int i = 0;  i < arraysize( _bits );  i++ )
         if ( _bits[ i ] != 0 )
            {
            uint64 data = _bits[ i ];
            int bit = i << 6;
            __asm
               {
               bsf rax, data
               add bit, eax
               }
            return bit;
            }
      // all slots of this vector are equal to 0
      return GetLength();
      }

   uint64 _bits[ 6 ];
   uint64 _filler[ 2 ]; // so that sizeof( BinVec384 ) is divisible by 64
   };

__declspec( align( 16 ) ) struct BinVec1280
   {
   BinVec1280() { memset( _bits, 0, sizeof( _bits ) ); }
   int GetLength() const { return sizeof( _bits ) * 8; }
   void SetAllBitsTo1() { memset( _bits, 0xFF, sizeof( _bits ) ); }
   void And( BinVec1280 const& a, BinVec1280 const& b )
      {
      __asm
         {
         mov      rbx, this
         mov      rsi, a
         mov      rdi, b
         movdqa   xmm0, [rsi]
         movdqa   xmm1, [rsi + 10h]
         movdqa   xmm2, [rsi + 20h]
         movdqa   xmm3, [rsi + 30h]
         pand     xmm0, [rdi]
         pand     xmm1, [rdi + 10h]
         pand     xmm2, [rdi + 20h]
         pand     xmm3, [rdi + 30h]
         movdqa   [rbx], xmm0
         movdqa   [rbx + 10h], xmm1
         movdqa   [rbx + 20h], xmm2
         movdqa   [rbx + 30h], xmm3
         movdqa   xmm4, [rsi + 40h]
         movdqa   xmm5, [rsi + 50h]
         movdqa   xmm6, [rsi + 60h]
         movdqa   xmm7, [rsi + 70h]
         pand     xmm4, [rdi + 40h]
         pand     xmm5, [rdi + 50h]
         pand     xmm6, [rdi + 60h]
         pand     xmm7, [rdi + 70h]
         movdqa   [rbx + 40h], xmm4
         movdqa   [rbx + 50h], xmm5
         movdqa   [rbx + 60h], xmm6
         movdqa   [rbx + 70h], xmm7
         movdqa   xmm8, [rsi + 80h]
         movdqa   xmm9, [rsi + 90h]
         pand     xmm8, [rdi + 80h]
         pand     xmm9, [rdi + 90h]
         movdqa   [rbx + 80h], xmm8
         movdqa   [rbx + 90h], xmm9
         }
      }
   bool AndAllZero( BinVec1280 const& a, BinVec1280 const& b ) const
      {
      int ret = 0;
      __asm
         {
         mov      rsi, a
         mov      rdi, b
         movdqa   xmm0, [rsi]
         movdqa   xmm1, [rsi + 10h]
         movdqa   xmm2, [rsi + 20h]
         movdqa   xmm3, [rsi + 30h]
         pand     xmm0, [rdi]
         pand     xmm1, [rdi + 10h]
         pand     xmm2, [rdi + 20h]
         pand     xmm3, [rdi + 30h]
         movdqa   xmm4, [rsi + 40h]
         movdqa   xmm5, [rsi + 50h]
         movdqa   xmm6, [rsi + 60h]
         movdqa   xmm7, [rsi + 70h]
         pand     xmm4, [rdi + 40h]
         pand     xmm5, [rdi + 50h]
         pand     xmm6, [rdi + 60h]
         pand     xmm7, [rdi + 70h]
         movdqa   xmm8, [rsi + 80h]
         movdqa   xmm9, [rsi + 90h]
         pand     xmm8, [rdi + 80h]
         pand     xmm9, [rdi + 90h]
         por      xmm0, xmm1
         por      xmm2, xmm3
         por      xmm4, xmm5
         por      xmm6, xmm7
         por      xmm8, xmm9
         por      xmm0, xmm2
         por      xmm4, xmm6
         por      xmm0, xmm8
         por      xmm0, xmm4
         ptest    xmm0, xmm0
         jnz      done
         mov      ret, 1
         done:
         }
      return ret != 0;
      }
   uint64 _bits[ 20 ];
   };

__declspec( align( 16 ) ) struct BinVec1536
   {
   BinVec1536() { memset( _bits, 0, sizeof( _bits ) ); }
   void SetAllBitsTo1() { memset( _bits, 0xFF, sizeof( _bits ) ); }
   int GetLength() const { return sizeof( _bits ) * 8; }
   void And( BinVec1536 const& a, BinVec1536 const& b )
      {
      __asm
         {
         mov      rbx, this
         mov      rsi, a
         mov      rdi, b
         movdqa   xmm0, [rsi]
         movdqa   xmm1, [rsi + 10h]
         movdqa   xmm2, [rsi + 20h]
         movdqa   xmm3, [rsi + 30h]
         movdqa   xmm4, [rsi + 40h]
         movdqa   xmm5, [rsi + 50h]
         movdqa   xmm6, [rsi + 60h]
         movdqa   xmm7, [rsi + 70h]
         pand     xmm0, [rdi]
         pand     xmm1, [rdi + 10h]
         pand     xmm2, [rdi + 20h]
         pand     xmm3, [rdi + 30h]
         pand     xmm4, [rdi + 40h]
         pand     xmm5, [rdi + 50h]
         pand     xmm6, [rdi + 60h]
         pand     xmm7, [rdi + 70h]
         movdqa   [rbx], xmm0
         movdqa   [rbx + 10h], xmm1
         movdqa   [rbx + 20h], xmm2
         movdqa   [rbx + 30h], xmm3
         movdqa   [rbx + 40h], xmm4
         movdqa   [rbx + 50h], xmm5
         movdqa   [rbx + 60h], xmm6
         movdqa   [rbx + 70h], xmm7
         lea rsi, [rsi + 80h]
         lea rdi, [rdi + 80h]
         lea rbx, [rbx + 80h]
         movdqa   xmm0, [rsi]
         movdqa   xmm1, [rsi + 10h]
         movdqa   xmm2, [rsi + 20h]
         movdqa   xmm3, [rsi + 30h]
         pand     xmm0, [rdi]
         pand     xmm1, [rdi + 10h]
         pand     xmm2, [rdi + 20h]
         pand     xmm3, [rdi + 30h]
         movdqa   [rbx], xmm0
         movdqa   [rbx + 10h], xmm1
         movdqa   [rbx + 20h], xmm2
         movdqa   [rbx + 30h], xmm3
         }
      }
   bool AndAllZero( BinVec1536 const& a, BinVec1536 const& b )
      {
      int ret = 0;
      __asm
         {
         mov      rsi, a
         mov      rdi, b
         movdqa   xmm0, [rsi]
         movdqa   xmm1, [rsi + 10h]
         movdqa   xmm2, [rsi + 20h]
         movdqa   xmm3, [rsi + 30h]
         pand     xmm0, [rdi]
         pand     xmm1, [rdi + 10h]
         pand     xmm2, [rdi + 20h]
         pand     xmm3, [rdi + 30h]
         por      xmm0, xmm1
         por      xmm2, xmm3
         por      xmm0, xmm2
         ptest    xmm0, xmm0
         jnz      done
         movdqa   xmm4, [rsi + 40h]
         movdqa   xmm5, [rsi + 50h]
         movdqa   xmm6, [rsi + 60h]
         movdqa   xmm7, [rsi + 70h]
         pand     xmm4, [rdi + 40h]
         pand     xmm5, [rdi + 50h]
         pand     xmm6, [rdi + 60h]
         pand     xmm7, [rdi + 70h]
         por      xmm4, xmm5
         por      xmm6, xmm7
         por      xmm4, xmm6
         ptest    xmm4, xmm4
         jnz      done
         lea rsi, [rsi + 80h]
         lea rdi, [rdi + 80h]
         movdqa   xmm0, [rsi]
         movdqa   xmm1, [rsi + 10h]
         movdqa   xmm2, [rsi + 20h]
         movdqa   xmm3, [rsi + 30h]
         pand     xmm0, [rdi]
         pand     xmm1, [rdi + 10h]
         pand     xmm2, [rdi + 20h]
         pand     xmm3, [rdi + 30h]
         por      xmm0, xmm1
         por      xmm2, xmm3
         por      xmm0, xmm2
         ptest    xmm0, xmm0
         jnz      done
         mov      ret, 1
         done:
         }
      return ret != 0;
      }

   uint64 _bits[ 24 ];
   };

#define LongBinVec( n, m )                                                             \
struct BinVec##n                                                                       \
   {                                                                                   \
   BinVec##n()                                                                         \
      {                                                                                \
      assert( n == m << 6 );                                                           \
      memset( _bits, 0, sizeof( _bits ) );                                             \
      }                                                                                \
   int GetLength() const { return sizeof( _bits ) * 8; }                               \
   void And( BinVec##n const& a, BinVec##n const& b )                                  \
      {                                                                                \
      for ( int i = 0;  i < arraysize( _bits );  i++ )                                 \
         _bits[ i ] = a._bits[ i ] & b._bits[ i ];                                     \
      }                                                                                \
   void SetAllBitsTo1() { memset( _bits, 0xFF, sizeof( _bits ) ); }                    \
   void SetBit( int bit )                                                              \
      {                                                                                \
      assert( bit >= 0  &&  bit < n );                                                 \
      _bits[ bit >> 6 ] |= (static_cast< uint64 >( 1 ) << (bit & 0x3F));               \
      }                                                                                \
   bool IsBitSet( int bit ) const                                                      \
      {                                                                                \
      assert( bit >= 0  &&  bit < n );                                                 \
      return (_bits[ bit >> 6 ] & (static_cast< uint64 >( 1 ) << (bit & 0x3F))) != 0;  \
      }                                                                                \
   uint64 _bits[ m ];                                                                  \
   }
   
LongBinVec( 8192, 128 );
LongBinVec( 16384, 256 );
LongBinVec( 32768, 512 );

#endif
